"""Translation of IDs with flexible formatting and name matching.

For an introduction to translation, see :ref:`translation-primer` and :ref:`mapping-primer`.


Environment variables
---------------------
.. envvar:: ID_TRANSLATION_DISABLED

   Global switch. When ``true``, the :meth:`Translator.translate`-method returns immediately.

.. envvar:: ID_TRANSLATION_SUPPRESS_OPTIONAL_FETCHER_INIT_ERRORS

   Global switch. Set to ``true`` (not recommended) to allow the :meth:`~id_translation.toml.TranslatorFactory` to
   discard `optional` fetchers that raise when imported or initialized.
   See the `documentation <../documentation/translator-config.html#optional-fetchers>`__ for details.
"""

import logging as _logging

from ._translator import Translator

__all__ = [
    "Translator",
    "__version__",  # Make MyPy happy
]

__version__ = "0.14.1"

_logging.getLogger(__name__).addHandler(_logging.NullHandler())
del _logging



from collections.abc import Mapping, Sequence
from typing import Any, Unpack

import pandas as pd

from ..offline.types import PlaceholderTranslations, SourcePlaceholderTranslations
from ..translator_typing import AbstractFetcherParams
from ..types import ID, IdType, SourceType
from ._abstract_fetcher import AbstractFetcher
from .types import FetchInstruction


class MemoryFetcher(AbstractFetcher[SourceType, IdType]):
    """Fetch from memory.

    Args:
        data: A dict ``{source: PlaceholderTranslations}`` to fetch from.
        return_all: If ``False``, return only the requested IDs and placeholders.
        **kwargs: See :class:`.AbstractFetcher`.
    """

    def __init__(
        self,
        data: (
            SourcePlaceholderTranslations[SourceType]
            | dict[SourceType, PlaceholderTranslations[SourceType]]
            | dict[SourceType, pd.DataFrame]
            | Mapping[SourceType, Mapping[str, Sequence[Any]]]
            | None
        ) = None,
        return_all: bool = True,
        **kwargs: Unpack[AbstractFetcherParams[SourceType, IdType]],
    ) -> None:
        super().__init__(**kwargs)
        self._data: SourcePlaceholderTranslations[SourceType] = (
            {} if not data else {source: PlaceholderTranslations.make(source, pht) for source, pht in data.items()}
        )
        self.return_all = return_all

    def _initialize_sources(self, task_id: int) -> dict[SourceType, list[str]]:  # noqa: ARG002
        return {source: list(pht.placeholders) for source, pht in self._data.items()}

    def fetch_translations(self, instr: FetchInstruction[SourceType, IdType]) -> PlaceholderTranslations[SourceType]:
        ret = self._data[instr.source]

        if self.return_all:
            return ret

        placeholder_indices = [
            ret.placeholders.index(placeholder)
            for placeholder in instr.placeholders  # requested
            if placeholder in ret.placeholders  # available
        ]

        if instr.ids is None:
            records = [[record[i] for i in placeholder_indices] for record in ret.records]
        else:
            ids = set(instr.ids)
            records = [
                [record[i] for i in placeholder_indices]
                for record in ret.records
                if record[ret.id_pos] in ids  # crash on missing IDs
            ]
        placeholders = tuple(ret.placeholders[i] for i in placeholder_indices)
        id_pos = placeholders.index(ID) if ID in placeholders else -1
        return PlaceholderTranslations(instr.source, placeholders, records, id_pos)



import logging
from typing import Any

from rics.misc import get_public_module, tname

from ..types import IdType, NameType, SourceType, TranslatableT
from ._data_structure_io import DataStructureIO
from ._dict import DictIO
from ._pandas import PandasIO
from ._sequence import SequenceIO
from ._set import SetIO
from ._single_value import SingleValueIO
from .exceptions import UntranslatableTypeError

AnyDataStructureIO = DataStructureIO[Any, Any, Any, Any]

_DEFAULT_IMPLEMENTATIONS: list[type[AnyDataStructureIO]] = [
    DictIO,
    SetIO,
    PandasIO,
    SequenceIO,
    SingleValueIO,
]
_RESOLUTION_ORDER: list[type[AnyDataStructureIO]] = [*_DEFAULT_IMPLEMENTATIONS]
LOGGER = logging.getLogger(__package__)


_ENTRYPOINT_GROUP: str = "id_translation.dio"
"""Alias of the ``__init__.ENTRYPOINT_GROUP`` attribute.

Background:
    https://github.com/sphinx-doc/sphinx/issues/6495#issuecomment-1058033697
    https://github.com/sphinx-doc/sphinx/issues/12020
"""


def resolve_io(
    arg: TranslatableT,
    **kwargs: Any,
) -> DataStructureIO[TranslatableT, NameType, SourceType, IdType]:
    """Get an IO instance for `arg`.

    Args:
        arg: An argument to get IO for.
        **kwargs: Keyword arguments for the IO class.

    Returns:
        A data structure IO instance for `arg`.

    Raises:
        UntranslatableTypeError: If no suitable IO implementation could be found.

    See Also:
        The :func:`register_io` function.
    """
    for io_class in _RESOLUTION_ORDER:
        if io_class.handles_type(arg):
            return _initialize(arg, io_class, kwargs)

    raise UntranslatableTypeError(type(arg))


def _initialize(
    arg: TranslatableT,
    io_class: type[DataStructureIO[TranslatableT, NameType, SourceType, IdType]],
    kwargs: dict[str, Any],
) -> DataStructureIO[TranslatableT, NameType, SourceType, IdType]:
    try:
        io = io_class(**kwargs)
        with_kwargs = True
    except TypeError:
        io = io_class()
        with_kwargs = False
    if LOGGER.isEnabledFor(logging.DEBUG):
        from rics.misc import format_kwargs

        pretty_arg = tname(arg, prefix_classname=True)
        pretty_class = f"{_pretty_io_name(io_class)}({format_kwargs(kwargs)})" if with_kwargs else ""
        index = _RESOLUTION_ORDER.index(io_class)
        LOGGER.debug(f"Using rank-{index} implementation {pretty_class} for translatable of type='{pretty_arg}'.")
    return io


def get_resolution_order(*, real: bool = False) -> list[type[AnyDataStructureIO]]:
    """Returns known :class:`.DataStructureIO` implementations.

    Args:
        real: If ``True``, return the actual list instead of a copy.

    Returns:
        A list of IO implementations.
    """
    return _RESOLUTION_ORDER if real else list(_RESOLUTION_ORDER)


def register_io(io: type[AnyDataStructureIO]) -> None:
    """Register a new IO implementation.

    Classes are polled through :meth:`.DataStructureIO.handles_type` in reverse insertion order (new implementations are
    polled first). Re-registering an implementation again will move it to the first position in the search order.

    Args:
        io: A :class:`.DataStructureIO` type.
    """
    if io in _RESOLUTION_ORDER:
        _RESOLUTION_ORDER.remove(io)

    _RESOLUTION_ORDER.insert(0, io)
    if LOGGER.isEnabledFor(logging.DEBUG):
        LOGGER.debug(f"Registered custom IO implementation: '{_pretty_io_name(io)}'.")


def _pretty_io_name(io: type[AnyDataStructureIO]) -> str:
    return get_public_module(io, resolve_reexport=True) + "." + tname(io, prefix_classname=True)


def is_registered(io: type[AnyDataStructureIO]) -> bool:
    """Return IO implementation registration status.

    Instances should register themselves using :func:`register_io` or :meth:`.DataStructureIO.is_registered`.

    Args:
        io: A :class:`.DataStructureIO` type.
    """
    return io in _RESOLUTION_ORDER


def load_integrations() -> None:
    """Discover, load and register entrypoint integrations.

    Reset the registry, then load entrypoints in the
    :const:`{_ENTRYPOINT_GROUP!r} <id_translation.dio.ENTRYPOINT_GROUP>`
    entrypoint group (see :py:func:`importlib.metadata.entry_points` for details).

    Will skip integrations that raise :class:`ImportError` when loaded.

    Raises:
        TypeError: If an integration does not inherit from :class:`.DataStructureIO`.
    """
    from importlib.metadata import entry_points
    from time import perf_counter

    global _RESOLUTION_ORDER  # noqa: PLW0603

    start = perf_counter()

    LOGGER.debug("Initializing %s integrations in group='%s'.", DataStructureIO.__name__, _ENTRYPOINT_GROUP)
    _RESOLUTION_ORDER = [*_DEFAULT_IMPLEMENTATIONS]

    n_total = 0
    n_ok = 0
    for ep in entry_points(group=_ENTRYPOINT_GROUP):
        n_total += 1

        try:
            cls = ep.load()
        except ImportError as e:
            LOGGER.debug("Failed to import entrypoint=%r: %r.", ep, e)
            continue

        if not issubclass(cls, DataStructureIO):
            msg = f"Bad entrypoint={ep!r}: {cls} is not a subtype of {DataStructureIO.__name__}. "
            raise TypeError(msg)

        cls.register()
        n_ok += 1

    millis = round(1000 * (perf_counter() - start))
    LOGGER.debug("Imported and registered %i/%i integrations in %i milliseconds.", n_ok, n_total, millis)


if load_integrations.__doc__:
    load_integrations.__doc__ = load_integrations.__doc__.format(_ENTRYPOINT_GROUP=_ENTRYPOINT_GROUP)



"""Global translation settings."""

import logging as _l
import typing as _t


class KeyEventLogLevel(_t.NamedTuple):
    """Enter/exit log level pair for key events. Default level is 10 ``logging.DEBUG=10``."""

    enter: int = _l.DEBUG
    """Log level for the ``ENTER`` message, e.g. ``TRANSLATOR.TRANSLATE.ENTER``.

    .. code-block:: python
       :caption: Example: Enter message of the :meth:`.Translator.translate`-method.

       Begin translation of 'DataFrame'-type data. Names to translate: Derive
         based on type='DataFrame'.
    """

    exit: int = _l.DEBUG
    """Log level for the ``EXIT`` message, e.g. ``TRANSLATOR.TRANSLATE.EXIT``.

    .. code-block:: python
       :caption: Example: Exit message of the :meth:`.Translator.translate`-method.

       Finished translation of 4 names in 'DataFrame'-type data in 133 ms,
         using name-to-source mapping: {'customer_id': 'customer', 'film_id':
         'film', 'category_id': 'category', 'staff_id': 'staff'}.
    """


class logging:  # noqa: N801
    """Global logging settings used by all instances."""

    TRANSLATE_ONLINE: KeyEventLogLevel = KeyEventLogLevel(exit=_l.INFO)
    """Levels for ``TRANSLATOR.TRANSLATE`` key event messages when the :class:`.Translator` is :attr:`~.Translator.online`."""
    TRANSLATE_OFFLINE: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``TRANSLATOR.TRANSLATE`` key event messages when the :class:`.Translator` is offline."""
    MAP: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``TRANSLATOR.MAP`` key event messages."""

    MAP_PLACEHOLDERS: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``FETCHER.MAP_PLACEHOLDERS`` key event messages."""
    FETCH_TRANSLATIONS: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``FETCHER.FETCH_TRANSLATIONS`` key event messages."""

    MULTI_FETCH: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``MULTIFETCHER.FETCH`` key event messages."""
    MULTI_FETCH_ALL: KeyEventLogLevel = KeyEventLogLevel()
    """Levels for ``MULTIFETCHER.FETCH_ALL`` key event messages."""

    def __init__(self) -> None:
        _raise_info_message(self)


def _raise_info_message(obj: _t.Any) -> None:
    from rics.misc import get_public_module, tname

    raise RuntimeError(
        f"Class '{get_public_module(obj)}.{tname(obj)}' is used as a public"
        f" namespace. There is no need to instantiate this class."
    )



"""Data structure IO exceptions."""

from typing import Any as _Any


class DataStructureIOError(TypeError):
    """Base class for IO exceptions."""


class UntranslatableTypeError(DataStructureIOError):
    """Exception indicating that a type cannot be translated.

    Args:
        t: A type.
    """

    def __init__(self, t: type[_Any]) -> None:
        super().__init__(
            f"Type {t} cannot be translated."
            "\nHint: https://id-translation.readthedocs.io/en/stable/api/id_translation.dio.html#user-defined-integrations"
        )


class NotInplaceTranslatableError(DataStructureIOError):
    """Exception indicating that a type cannot be translated in-place.

    Args:
        arg: Something that can't be translated inplace.
    """

    def __init__(self, arg: _Any) -> None:
        super().__init__(
            f"Inplace translation not possible or implemented for type: {type(arg)}"
            "\nHint: https://id-translation.readthedocs.io/en/stable/api/id_translation.dio.html#user-defined-integrations"
        )



"""Logging utilities."""

import typing as _t
from uuid import UUID as _UUID

import numpy as _np


def cast_unsafe(obj: _t.Any) -> _t.Any:
    """Attempt to cast an arbitrary object to a JSON-safe type."""
    for clazz, formatter in FORMATTERS.items():
        if isinstance(obj, clazz):
            return formatter(obj)
    return obj


FORMATTERS: dict[type[_t.Any], _t.Any] = {
    _UUID: str,
    _np.integer: int,
    _np.floating: float,
    _np.ndarray: _np.ndarray.tolist,
    list: lambda x: [cast_unsafe(obj) for obj in x],
}
"""Formatters used to cast unsafe JSON types."""
